/**
 * Copyright (c) 2010-2012 LAAS/CNRS
 * Arnaud Degroote - Wed Aug 11 2010
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#ifndef _TCLSERV_CLIENT_H_
#define _TCLSERV_CLIENT_H_

#if defined(__GNUC__) || defined(__clang__)
#define DEPRECATED(func) func __attribute__ ((deprecated))
#elif defined(_MSC_VER)
#define DEPRECATED(func) __declspec(deprecated) func
#else
#pragma message("WARNING: You need to implement DEPRECATED for this compiler")
#define DEPRECATED(func) func
#endif


#ifdef __cplusplus
extern "C" {
#endif

#include <stdbool.h>

typedef struct tclserv_client* TCLSERV_CLIENT_ID;

/**
 * Initialize a connection to a tclserv
 *
 * @param host is canonical name of host
 * @param port can be a port number or a name service
 * @return NULL in case of error, otherwise a valid TCLSERV_CLIENT_ID
 */ 
TCLSERV_CLIENT_ID tclserv_client_init(const char* host, const char* port);

/**
 * Initialize a connection to a tclserv
 *
 * @param host is canonical name of host
 * @param port can be a port number or a name service
 * Normally, we use an internal thread to simulate asynchronism. However, when
 * integrating in larger software, you may already have already some event
 * loop, or other ways to do some asynchronous stuff. If it is the case, just
 * set @param use_internal_thread to false. You can use the method
 * @fn tclserv_client_get_answer_socket to retrieve a base descriptor for answer
 * socket and call @fn tclserv_client_process_answer_msg when new datas arrives
 * on it.
 * @return NULL in case of error, otherwise a valid TCLSERV_CLIENT_ID
 */ 
TCLSERV_CLIENT_ID tclserv_client_init_ex(const char* host, const char* port,
									     bool use_internal_thread);

/**
 * Close a connection to a remote tclserv and release all ressources 
 */
void tclserv_client_destroy(TCLSERV_CLIENT_ID*);


/**
 * @return a file descriptor we can use to wait for new msg
 */
int tclserv_client_get_answer_socket(TCLSERV_CLIENT_ID id);

/**
 * Wait for a new message and process it
 *
 * @param id is a valid TCLSERV_CLIENT_ID
 * @param rqstId will store, if non-NULL, the rqstId associated with the processed msg 
 *
 * @return -1 if we fails to process the message, in thise case, rqstId is in
 * an unknow state, otherwise 0.
 */
int tclserv_client_process_answer_msg(TCLSERV_CLIENT_ID id, ssize_t* rqstId);


/**
 * Wait for a new message and process it
 *
 * @param id is a valid TCLSERV_CLIENT_ID
 * @param rqstId will store, if non-NULL, the rqstId associated with the processed msg 
 * @param tv is the time is the maximum time we wait for a new message. If tv is NULL, 
 * the function has the same semantic than tclserv_client_process_answer_msg
 *
 * @return -1 if we fails to process the message or if we got a timeout (errno
 * == ETIME), in thise case, rqstId is in an unknow state, otherwise 0.
 */
int tclserv_client_timed_process_answer_msg(TCLSERV_CLIENT_ID id, ssize_t* rqstId, struct timeval* tv);

/**
 * Load a module on a remote server
 * @param id is valid TCLSERV_CLIENT_ID
 * @param module is a module name
 *
 * This function is deprecated, you must use tclserv_client_load_module_ex
 * Keep it for compatibility with code generated by genom
 *
 * @return -1 in case of failure, 0 otherwise
 */
DEPRECATED(int tclserv_client_load_module(TCLSERV_CLIENT_ID id, const char* module));

/**
 * Load a module on a remote server
 * @param id is valid TCLSERV_CLIENT_ID
 * @param module is a module name
 * @param why contains the error msg if any. If NULL, don't care about the error msg
 *
 * @return -1 in case of failure, 
 *			 0 otherwise
 */
int tclserv_client_load_module_ex(TCLSERV_CLIENT_ID id, const char* module, char** why);

/**
 * Unload a module on a remote server
 * @param id is valid TCLSERV_CLIENT_ID
 * @param module is a module name
 *
 * @return -1 in case of failure, 0 otherwise
 */
int tclserv_client_unload_module(TCLSERV_CLIENT_ID id, const char* module);

/**
 * Send a request to a remote server
 * @param id is a valid TCLSERV_CLIENT_ID
 * @param rqst is a valid request like pom::SetModel pipo (don't need to be
 * terminated by \r\n"
 *
 * @return -1 in failure case
 * Otherwise @return the rqstId associated to this request
 */
ssize_t tclserv_client_rqst(TCLSERV_CLIENT_ID id, const char* rqst);


/**
 * Send a request to a remote server and wait for ack
 * @param id is a valid TCLSERV_CLIENT_ID
 * @param rqst is a valid request like pom::SetModel pipo (don't need to be
 * terminated by \r\n"
 *
 * @return -1 in failure case
 * Otherwise @return the rqstId associated to this request
 */
ssize_t tclserv_client_rqst_ack(TCLSERV_CLIENT_ID id, const char* rqst);

/**
 * Check if request is terminated (for asynchronous call)
 * @param id is a valid TCLSERV_CLIENT_ID
 * @param rqstId is a valid rqstId (received by rqst or rqst_ask)
 *
 * @return -2 in failure case
 * @return -1 if the task has not been terminated
 * @return 0 if the task has terminated
 */
int tclserv_client_has_terminated(TCLSERV_CLIENT_ID id, ssize_t rqstId);

/**
 * Wait for the terminaison of a rqst
 * @param id is a valid TCLSERV_CLIENT_ID
 * @param rqstId is a valid rqstId (received by rqst or rqst_ask)
 * @param ans will store the output of the server. It must be freed by the caller.
 * If @param ans is NULL, nothing is stored, you will only got if it is success or
 * not.
 *
 * @return -2 in failure case
 * @return -1 if the request has terminated with an error. @param ans contains the
 * error msg
 * @return 0 if the task has terminated with success. @param ans contains OK followed
 * by a serialized output of the rqst
 */
int tclserv_client_wait(TCLSERV_CLIENT_ID id, ssize_t rqstId, char** ans);

/**
 * Send a request to a remote server and wait for the answer
 * @param id is a valid TCLSERV_CLIENT_ID
 * @param rqst is a valid request like pom::SetModel pipo (don't need to be
 * terminated by \r\n"
 * @param ans will store the output of the server. It must be freed by the caller.
 *
 * @return -2 in failure case
 * @return -1 if the request has terminated with an error. @param ans contains the
 * error msg
 * @return 0 if the task has terminated with success. @param ans contains OK followed
 * by a serialized output of the rqst
 */
int tclserv_client_rqst_wait(TCLSERV_CLIENT_ID id, const char* rqst, char** ans);

/**
 * Try to abort a remote task
 * @param id is a valid TCLSERV_CLIENT_ID
 * @param rqstId is a valid request Id
 * @param wait notes if we wait for the answer on answer channel
 *
 * @return -3 if we are in initial mode (i.e, not yet acked) and wait == false.
 * In this case, the caller must retry later.
 * @return -2 in failure case
 * @return -1 if the msg has been send, but the server can't abort the request
 * @return 0 if wait == true and the msg has been send and the request has been aborted
 * @return a valid rqstId if wait == false and the msg has been send and the
 * request has been aborted
 */
ssize_t tclserv_client_abort_ex(TCLSERV_CLIENT_ID id, ssize_t rqstId, bool wait);

/**
 * Deprecated
 * Equivalent to tclserv_client_abort(id, rqstId, true)
 */
DEPRECATED(int tclserv_client_abort(TCLSERV_CLIENT_ID id, ssize_t rqstId));

/**
 * List the modules loaded on the remote tclserv
 * @param id is a valid TCLSERV_CLIENT_ID
 * @param answer will store the output of the server. It must be freed by the caller
 *
 * @return -2 in bad failure case
 * @return -1 if we successfully send the request, but the server returns an error
 * @return 0 in success case. *answer will be the list of modules, separated by space
 */
int tclserv_client_list_modules(TCLSERV_CLIENT_ID id, char** answer); 

/**
 * Execute a command on the remote server
 * @param id is a valid TCLSERV_CLIENT_ID
 * @param rqst is the command that will be executed on the remote host
 * @param answer will store the answer msg if any. It must be freed by the caller
 *
 * @return -2 in bad failure case
 * @return -1 if we successfully send the cmd, but the sever returns an error
 * @return 0 in success case. answer contains the undecoded part of the answer.
 */
int tclserv_client_exec(TCLSERV_CLIENT_ID id, const char* rqst, char** answer);

/**
 * Start a module on the remote server (effectivly start the processus module)
 * @param id is a valid TCLSERV_CLIENT_ID
 * @param module is the name of module you want to start
 *
 * This will be deprecated for tclserv_client_start_module_ex. Call it with NULL
 * if you don't care about the reason failure
 *
 * @return -2 in bad failure case
 * @return -1 if we succesfullay ask the server to start the module
 * @return the pid of the process in success case (> 0).
 */
DEPRECATED(int tclserv_client_start_module(TCLSERV_CLIENT_ID id, const char* module));

/**
 * Start a module on the remote server (effectivly start the processus module)
 * @param id is a valid TCLSERV_CLIENT_ID
 * @param module is the name of module you want to start
 * @param why will store the error if not NULL. It must be freed by user
 *
 * @return -2 in bad failure case
 * @return -1 if we succesfullay ask the server to start the module, but the
 * server returns an error. The error is printed on stderr
 * @return the pid of the process in success case (> 0).
 */
int tclserv_client_start_module_ex(TCLSERV_CLIENT_ID id, const char* module, char**why);

/**
 * Stop a module on the remote server
 *
 * To do it, we currently use the KILL method from tclserv. Howerver, it
 * supposes that not only the module is started, but it is also loaded (using
 * LM).  It is not really symetric with start_module, so in the future, we may
 * call directly killmodule
 *
 * @param id is a valid TCLSERV_CLIENT_ID
 * @param module is the name of module you want to stop
 *
 * The function is deprecated for tclserv_client_stop_module_ex.
 *
 * @return -2 in bad failure case
 * @return -1 if we succesfullay ask the server to start the module, but the
 * server returns an error. 
 * @return 0 in success case.
 */ 
DEPRECATED(int tclserv_client_stop_module(TCLSERV_CLIENT_ID id, const char* module));

/**
 * Same as previous, but why stores the error msg if any. In this case, it must
 * be freed by user
 */
int tclserv_client_stop_module_ex(TCLSERV_CLIENT_ID id, const char* module, char **why);


#ifdef __cplusplus
}
#endif
#endif /* _TCLSERV_CLIENT_H_ */
